#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh
# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/longhornutils.sh

verify_longhorn_nodes() {
  ds_count=$(kubectl get daemonsets.apps longhorn-manager  -n longhorn-system -o jsonpath='{.status.desiredNumberScheduled}')
  node_count=
  for _ in {1..10}; do
    node_count=$(kubectl get nodes.longhorn.io -n longhorn-system -o go-template='{{len .items}}')
    [[ ${node_count} -eq ${ds_count} ]] && break
    sleep 60
  done
  [[ ${node_count} -ne ${ds_count} ]] && warn "Longhorn nodes($((ds_count - node_count))) not deployed yet" && return 1

  IFS=" " read -r -a node_list <<< "$(kubectl get nodes.longhorn.io -n longhorn-system -o jsonpath='{.items..spec.name}')"
  ready=0
  ready_node_count=0
  for node in "${node_list[@]}"; do
    ready=0
    for _ in {1..10}; do
      if [ "$(kubectl get nodes.longhorn.io/"${node}" -n longhorn-system -o json | jq -r 'try (.status.conditions[] | select(.type=="Ready").status)')" == "True" ]; then
        ready=1 && ready_node_count=$((ready_node_count+1)) && break
      fi
      sleep 60
    done
    [[ ${ready} -ne 1 ]] && warn "Longhorn node ${node} is not ready"
  done
  [[ ${ready_node_count} -ne ${node_count} ]] && warn "Longhorn nodes($((node_count - ready_node_count))) not ready yet" && return 1
  info "Longhorn nodes are ready" && return 0
}

verify_longhorn_instance_manager() {
  IFS=" " read -r -a node_list <<< "$(kubectl get nodes.longhorn.io -n longhorn-system -o jsonpath='{.items..spec.name}')"
  node_count=${#node_list[@]}
  deployed_node_count=0
  ready=0
  maxtry=5
  for node in "${node_list[@]}"; do
    ready=0
    for instance in engine replica; do
      if [ "$instance" == "replica" ]; then
        local try=0
        local has_disk

        while (( try != maxtry )); do
          has_disk=$(curl -s -X GET --noproxy "*" "${LONGHORN_URL}/v1/nodes/${node,,}" |jq -r '.disks!=null and .disks!={}') || true
          [[ -n "${has_disk}" ]] && break
          sleep 60
          try=$((try+1))
        done
        [[ -z "${has_disk}" ]] && warn "Longhorn backend is not responding" && return 1
	      [[ ${has_disk} == "false" ]] && info "Skipping replica check for ${node}" && continue
      fi
      ready=0
      for _ in {1..10}; do
	status=$(kubectl get instancemanagers -n longhorn-system -l "longhorn.io/instance-manager-type=${instance},longhorn.io/node=${node}" --sort-by='.metadata.creationTimestamp' -o jsonpath='{.items[-1].status.currentState}')
        [[ "${status}" == "running" ]] && ready=1 && break
        sleep 60
      done
      [[ ${ready} -ne 1 ]] && warn "Longhorn instance manager ${instance} is not ready on node ${node}" && break
    done
    [[ ${ready} -eq 1 ]] && deployed_node_count=$((deployed_node_count+1))
  done
  [[ ${deployed_node_count} -ne ${node_count} ]] && warn "Longhorn instance managers are not ready on $((node_count - deployed_node_count)) nodes" && return 1
  info "Longhorn instance managers are deployed" && return 0
}

verify_longhorn_engineimages() {
  ready=0
  for _ in {1..10}; do
    if [ "$(kubectl get engineimages -n longhorn-system --sort-by=.metadata.creationTimestamp -o 'jsonpath={.items[-1].status.state}')" == "deployed" ]; then
      ready=1 && break
    fi
    sleep 60
  done
  [[ ${ready} -ne 1 ]] && warn "Longhorn engine images not deployed yet" && return 1
  info "Longhorn engine images are deployed" && return 0
}

main() {
  set_longhorn_url
  verify_longhorn_nodes || return 1
  verify_longhorn_instance_manager || return 1
  verify_longhorn_engineimages || return 1
  return 0
}

main