#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

LANG_FABRIC_FAILED_PREFIX="[ManagementTools] [Failed]"
LANG_STORAGE_SETUP_MESSAGE="Installing and configuring storage"

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/longhornutils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh


function load_istio_variables() {
  PROFILE="$(get_mandatory_json_value ".profile")"
  FAULT_TOLERANCE="$(get_json_value_or_return_default_value ".fault_tolerance" "1")"

  [[ "${PROFILE}" == "ha" && "$FAULT_TOLERANCE" -lt 1 ]] && error "Unsupported Fault tolerance value for HA profile. Expected: '>=1' Current: '${FAULT_TOLERANCE}'"
  [[ "${PROFILE}" != "ha" && "$FAULT_TOLERANCE" -ne 0 ]] && info "Setting Fault tolerance to '0' for non-HA profile" && FAULT_TOLERANCE=0

  DOCKER_REGISTRY_PULL_SECRET="$(get_optional_json_value ".registries.docker.pull_secret_value")"
  DOCKER_REGISTRY_URL="$(get_mandatory_json_value ".registries.docker.url")"
  PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
}

function updating_deprecated_api_mapkubeapis() {
  info "Executing helm status for istio-configure chart with namespace uipath-infra"
  existing_istio_configure_chart=$(helm status istio-configure -n uipath-infra || true)
  if [[ -n "${existing_istio_configure_chart}" ]]; then
    info "istio-configure release found in uipath-infra namespace"
    info "Updating the deprecated api in istio-configure"
    info "Executing helm mapkubeapis for istio-configure chart with namespace uipath-infra"
    helm mapkubeapis istio-configure -n uipath-infra
  fi

  existing_istio_chart=$(helm status istio -n istio-system || true)
  if [[ -n "${existing_istio_chart}" ]]; then
    info "Updating the deprecated api in istio"
    info "Executing helm mapkubeapis for istio chart with namespace istio-system"
    helm mapkubeapis istio -n istio-system
  fi
}

function delete_docker_vs(){
if [[ ${DO_UPGRADE} == "1" ]]; then
    if [[ ${INSTALL_TYPE} == "online" ]]; then
     kubectl delete destinationrule dockerregistry-dtrl -n istio-system --ignore-not-found
     kubectl delete  virtualservices dockerregistry-vs -n istio-system --ignore-not-found
  fi
fi
}


function annotate_docker_destination_rule(){
if [[ ${DO_UPGRADE} == "1" ]]; then
    if [[ ${INSTALL_TYPE} == "offline" ]]; then
      dtr=$(kubectl get destinationrule/dockerregistry-dtrl -n istio-system --ignore-not-found)
        if [[ -n "${dtr}" ]];
        then
          kubectl annotate destinationrule/dockerregistry-dtrl galley.istio.io/analyze-suppress=IST0128 -n istio-system
        fi 
    fi
fi

}

function install_istio() {
  info "Installing Istio"

  local namespace="istio-system"
  local ISTIOD_REPLICA_MIN_COUNT
  local ISTIOD_REPLICA_MAX_COUNT
  local PULL_SECRET_NAME="uipathpullsecret"
  local ENVOY_LUA_LIB_CM="envoy-lua-lib"
  local istio_helm_tmp_dir

  kubectl get ns ${namespace} &>>"${dev_null}" || kubectl create ns ${namespace} &>>"${dev_null}"

  declare -a ISTIO_CHART_ARGS

if [[ ${INSTALL_TYPE} == "online" ]]; then
    ISTIO_CHART_ARGS+=(--set global.cattle.systemDefaultRegistry="${DOCKER_REGISTRY_URL}")
  fi

  local image_pull_secret_tmp="../Infra_Installer/rancher-istio/image-pull-secret-tmp.yaml"
  local istio_values_overlay_tmp="../Infra_Installer/rancher-istio/values-overlay-tmp.yaml"

  \cp ../Infra_Installer/rancher-istio/image-pull-secret.yaml "${image_pull_secret_tmp}"
  \cp ../Infra_Installer/rancher-istio/values-overlay.yaml "${istio_values_overlay_tmp}"

  if [ -n "${DOCKER_REGISTRY_PULL_SECRET}" ]
  then
    sed -i "s;NAMESPACE_PLACEHOLDER;${namespace};"  "${image_pull_secret_tmp}"
    sed -i "s;IMAGE_PULL_SECRET_PLACEHOLDER;${PULL_SECRET_NAME};" "${image_pull_secret_tmp}"
    sed -i "s;IMAGE_PULL_SECRET_VALUE_PLACEHOLDER;${DOCKER_REGISTRY_PULL_SECRET};" "${image_pull_secret_tmp}"
    kubectl apply -f "${image_pull_secret_tmp}" &>>"${dev_null}"
    sed -i "s;LIST_IMAGE_PULL_SECRET_PLACEHOLDER;\[\"${PULL_SECRET_NAME}\"\];" "${istio_values_overlay_tmp}"
  else
    sed -i "s;LIST_IMAGE_PULL_SECRET_PLACEHOLDER;\[\];" "${istio_values_overlay_tmp}"
  fi

  if [[ "${PROXY_ENABLED}" == "true" ]] ; then
    local no_proxy="${NO_PROXY}",127.0.0.0/8,127.0.0.0,localhost,kubernetes,.svc,.local,.cluster
    sed -i "s;__NO_PROXY_PLACEHOLDER__;${no_proxy};g" "${istio_values_overlay_tmp}"
    sed -i "s;__HTTP_PROXY_PLACEHOLDER__;${HTTP_PROXY};g" "${istio_values_overlay_tmp}"
    sed -i "s;__HTTPS_PROXY_PLACEHOLDER__;${HTTPS_PROXY};g" "${istio_values_overlay_tmp}"
  else
    sed -i -E '/(name|value):.*_proxy/Id' "${istio_values_overlay_tmp}"
  fi

  if [[ "${PROFILE}" == "ha" ]]
  then
    ISTIOD_REQUESTS_CPU=500m
    ISTIOD_REQUESTS_MEMORY=2Gi
    ISTIOD_REPLICA_MIN_COUNT=2
    ISTIOD_REPLICA_MAX_COUNT=3
    ISTIO_PROXY_REQUESTS_CPU=50m
    ISTIO_PROXY_REQUESTS_MEMORY=128Mi
  else
    ISTIOD_REQUESTS_CPU=100m
    ISTIOD_REQUESTS_MEMORY=500Mi
    ISTIOD_REPLICA_MIN_COUNT=1
    ISTIOD_REPLICA_MAX_COUNT=1
    ISTIO_PROXY_REQUESTS_CPU=10m
    ISTIO_PROXY_REQUESTS_MEMORY=50Mi
  fi

  sed -i "s;NAMESPACE_PLACEHOLDER;${namespace};" "${istio_values_overlay_tmp}"
  sed -i "s;NAMESPACE_PLACEHOLDER;${namespace};" "${istio_values_overlay_tmp}"
  sed -i "s;ISTIOD_REQUESTS_CPU_PLACEHOLDER;${ISTIOD_REQUESTS_CPU};" "${istio_values_overlay_tmp}"
  sed -i "s;ISTIOD_REQUESTS_MEMORY_PLACEHOLDER;${ISTIOD_REQUESTS_MEMORY};" "${istio_values_overlay_tmp}"
  sed -i "s;ISTIOD_REPLICA_MIN_COUNT_PLACEHOLDER;${ISTIOD_REPLICA_MIN_COUNT};" "${istio_values_overlay_tmp}"
  sed -i "s;ISTIOD_REPLICA_MAX_COUNT_PLACEHOLDER;${ISTIOD_REPLICA_MAX_COUNT};" "${istio_values_overlay_tmp}"
  sed -i "s;ISTIO_PROXY_REQUESTS_CPU_PLACEHOLDER;${ISTIO_PROXY_REQUESTS_CPU};" "${istio_values_overlay_tmp}"
  sed -i "s;ISTIO_PROXY_REQUESTS_MEMORY_PLACEHOLDER;${ISTIO_PROXY_REQUESTS_MEMORY};" "${istio_values_overlay_tmp}"
  sed -i "s;ENVOY_LUA_LIB_CM_PLACEHOLDER;${ENVOY_LUA_LIB_CM};" "${istio_values_overlay_tmp}"

  kubectl -n ${namespace} create configmap "${ENVOY_LUA_LIB_CM}" \
    --from-file=../Infra_Installer/rancher-istio/lua-lib/src/  --dry-run=client -o yaml \
    | kubectl -n $namespace apply -f - &>>"${dev_null}"

  ## Delete istiod deployment to allow subsequent runs to pass
  ## Please follow this ticket for more update https://uipath.atlassian.net/browse/SERFAB-2071
 # if kubectl --namespace $namespace get deployment/istiod
 # then
 #   kubectl --namespace $namespace get deployment/istiod -o yaml > /tmp/istiod-deployment.yaml
 #   kubectl --namespace $namespace delete --wait=true deployment/istiod
 #   kubectl --namespace $namespace delete pods -l app=istio-ingressgateway
 # fi
  if [[ ${DO_UPGRADE} == "1" ]]; then
    updating_deprecated_api_mapkubeapis
    download_istio_helm_chart
    istio_helm_tmp_dir="${tmp_helm_charts_dir}"
    else
    istio_helm_tmp_dir="${TMP_HELM_CHARTS_DIR}"
  fi


  helm upgrade --atomic --install -f "${istio_values_overlay_tmp}" istio "${istio_helm_tmp_dir}"/rancher-istio/ \
    --wait --timeout 10m \
    --namespace $namespace \
    --create-namespace \
    --debug \
    "${ISTIO_CHART_ARGS[@]}" &>>"${dev_null}" || {
      test -f /tmp/istiod-deployment.yaml && kubectl apply -f /tmp/istiod-deployment.yaml;
      error_without_exit "Failed to install istio"
      error_with_code "11012";
    }

  info "Finished installing Istio"
}

function configure_istio() {
  info "Configuring Istio"

  local infra_namespace="uipath-infra"
  local istio_namespace="istio-system"
  local istio_tls_cert_name="istio-ingressgateway-certs"
  local istio_tls_alt_cert_name="istio-ingressgateway-alt-certs"
  local ca_cert_name="ca.crt"
  local tls_cert_name="tls.crt"
  local tls_key_name="tls.key"
  local label='config-discovery=yes'
  local istio_helm_tmp_dir
  local UPDATE_FQDN=$(get_cluster_config "update_fqdn" )
  local ACTIVE_ACTIVE="$(get_optional_json_value ".active_active.enabled")"

  secret=$(kubectl get secret "${istio_tls_cert_name}" \
    --namespace "${istio_namespace}" \
    --ignore-not-found \
    --no-headers)

  if [[ "${secret}" == "" ]] || [[ "${UPDATE_FQDN}" == "true" ]];
  then
    # only update certificate if not already configured or updating fqdn
    info "configuring certificate in istio / cluster"
    tmp_dir=$(mktemp -d certificates.XXXXXXX)
    tmp_file="ca_tls_merged.crt"
    add_empty_line_in_file_if_not_exists "${CERT_TLS_FILE_PATH}"

    cat "${CERT_TLS_FILE_PATH}" "${CERT_CA_FILE_PATH}" >> "${tmp_dir}"/"${tmp_file}"
    merge_ca_and_tls_path="${tmp_dir}"/"${tmp_file}"

    kubectl create secret generic "${istio_tls_cert_name}" \
      --from-file="${ca_cert_name}"="${CERT_CA_FILE_PATH}" \
      --from-file="${tls_cert_name}"="${merge_ca_and_tls_path}" \
      --from-file="${tls_key_name}"="${CERT_KEY_FILE_PATH}" \
      --namespace "${istio_namespace}" \
      --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"
    
    rm -rf "${tmp_dir}"
    
    kubectl label secret ${istio_tls_cert_name} "${label}" \
      --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"

    kubectl label secret ${istio_tls_cert_name} monitor-cert="true" \
      --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"
  fi

  if [[ -n "${ALTERNATIVE_FQDN}" ]]; then
    tmp_dir=$(mktemp -d certificates.XXXXXXX)
    alt_secret=$(kubectl get secret "${istio_tls_alt_cert_name}" \
      --namespace "${istio_namespace}" \
      --ignore-not-found \
      --no-headers -o jsonpath="{.data.tls\.crt}")
    dns_found="false"
    if [[ -n "${alt_secret}" ]]; then
      echo "${alt_secret}" | base64 -d > "${tmp_dir}/tls.crt"
      csplit -s -z -f "${tmp_dir}/pem-" "${tmp_dir}/tls.crt" '/-----BEGIN CERTIFICATE-----/' '{*}'
      tmp_dir_absolute_path=$(readlink -f "${tmp_dir}")
      for i in "${tmp_dir_absolute_path}"/pem-*; do
        info "Checking cert for SAN: ${i}"
        response=$(validate_certificate_san "${i}" "*.${ALTERNATIVE_FQDN}")
        [[ -z ${response} ]] && {
          dns_found="true"
          info "SAN configured in the cert"
          break
        }
      done
      rm -rf "${tmp_dir_absolute_path}"
    fi

    if [[ "${DO_UPGRADE}" == "0" ]] && [[ -z "${alt_secret}" || ("${UPDATE_FQDN}" == "true" || "${dns_found}" == "false" ) ]]; then
      # only update certificate if not already configured or updating fqdn
      info "configuring alt_fqdn certificate in istio / cluster"
      tmp_dir=$(mktemp -d certificates.XXXXXXX)
      tmp_file="ca_tls_merged.crt"
      cat "${ALT_CERT_TLS_FILE_PATH}" "${ALT_CERT_CA_FILE_PATH}" >> "${tmp_dir}"/"${tmp_file}"
      merge_ca_and_tls_path="${tmp_dir}"/"${tmp_file}"

      kubectl create secret generic "${istio_tls_alt_cert_name}" \
        --from-file="${ca_cert_name}"="${ALT_CERT_CA_FILE_PATH}" \
        --from-file="${tls_cert_name}"="${merge_ca_and_tls_path}" \
        --from-file="${tls_key_name}"="${ALT_CERT_KEY_FILE_PATH}" \
        --namespace "${istio_namespace}" \
        --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"

      rm -rf "${tmp_dir}"

      kubectl label secret ${istio_tls_alt_cert_name} "${label}" \
        --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"

      kubectl label secret ${istio_tls_alt_cert_name} monitor-cert="true" \
        --namespace ${istio_namespace} --overwrite=true &>>"${dev_null}"
    fi
  fi

  kubectl get namespace "${infra_namespace}" &>>"${dev_null}" \
    || kubectl create namespace "${infra_namespace}" &>>"${dev_null}" \
    || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
      error_with_code "10003" "${infra_namespace}";
    }

  declare -a ISTIO_CONFIGURE_CHART_ARGS

  ISTIO_CONFIGURE_CHART_ARGS+=(--set fqdn="${CLUSTER_FQDN}")
  # if CLUSTER_FQDN and ALTERNATIVE_FQDN are the same, don't pass ALTERNATIVE_FQDN to prevent duplicate hostnames.
  [[ "${CLUSTER_FQDN}" != "${ALTERNATIVE_FQDN}" ]] && ISTIO_CONFIGURE_CHART_ARGS+=(--set alt_fqdn="${ALTERNATIVE_FQDN}") || ISTIO_CONFIGURE_CHART_ARGS+=(--set alt_fqdn="")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.insightslooker.fqdn="${LOOKER_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.cephobjectstore.fqdn="${OBJECTSTORE_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.rancherserver.fqdn="${RANCHER_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.argocdserver.fqdn="${ARGO_SUBDOMAIN}")
  ISTIO_CONFIGURE_CHART_ARGS+=(--set ingresses.dockerregistry.fqdn="${DOCKER_REGISTRY_SUBDOMAIN}")

  if [[ ${INSTALL_TYPE} == "online" ]]; then
    ISTIO_CONFIGURE_CHART_ARGS+=(--set uipath.registry="${DOCKER_REGISTRY_URL}")
  else
    ISTIO_CONFIGURE_CHART_ARGS+=(--set uipath.registry="")
  fi

  if [[ ${INSTALLSTANDALONEAICENTER} == "1" ]]; then
      ISTIO_CONFIGURE_CHART_ARGS+=(--set orchestratorTrafficPolicies.enabled="false" )
      ISTIO_CONFIGURE_CHART_ARGS+=(--set firstPartyTrafficPolicies.enabled="false" )
      ISTIO_CONFIGURE_CHART_ARGS+=(--set identityTrafficPolicies.enabled="false" )
  fi

  if [[ ${ACTIVE_ACTIVE} == "true" ]]; then
      if [[ ${PRIMARY_CLUSTER} == "false" ]]; then
        [[ -z "${PRIMARY_CLUSTER_FQDN}" ]] && error_with_code "15013"
        ISTIO_CONFIGURE_CHART_ARGS+=(--set requestForwarding.target="${PRIMARY_CLUSTER_FQDN}" )
      fi
  fi

  [ -z "${DOCKER_REGISTRY_PULL_SECRET}" ] || ISTIO_CONFIGURE_CHART_ARGS+=(--set global.imagePullSecret.value="${DOCKER_REGISTRY_PULL_SECRET}")

  if [[ ${DO_UPGRADE} == "1" ]]; then
    updating_deprecated_api_mapkubeapis
    download_istio_configure_chart
    istio_helm_tmp_dir="${tmp_helm_charts_dir}"
  else
    #shellcheck disable=SC2154,SC2153
    istio_helm_tmp_dir="${TMP_HELM_CHARTS_DIR}"
  fi  

  helm upgrade --install --atomic istio-configure "${istio_helm_tmp_dir}"/istio-configure/ \
    --wait --timeout 10m \
    --namespace $infra_namespace \
    --create-namespace \
    "${ISTIO_CONFIGURE_CHART_ARGS[@]}" &>>"${dev_null}" || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_ROUTING_SETUP_MESSAGE}";
      error_with_code "11013";
    }

    info "Finished Configuring Istio"
}

function validate_subdomain_resolution() {
  local domain
  local domain_ip
  local pod_domain_ip
  local istio_namespace="istio-system"
  local error_messages=()
  local warn_messages=()
  local resolution_try=3
  local dns_validation_file_tmp="../Infra_Installer/rancher-istio/tests/dns_validation-tmp.yaml"

  \cp ../Infra_Installer/rancher-istio/tests/dns_validation.yaml "${dns_validation_file_tmp}"

  info "Verifying healthy status of istio ingress pods"
  kubectl -n istio-system rollout status --timeout=600s ds/istio-ingressgateway || error "Timeout waiting for istio ingress pods to become healthy"

  info "Creating dns validation test pods"
   if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;__REGISTRY_PLACEHOLDER__;${DOCKER_REGISTRY_URL}/;" "${dns_validation_file_tmp}"
  else
    sed -i "s;__REGISTRY_PLACEHOLDER__;;" "${dns_validation_file_tmp}"
  fi
  kubectl apply -f "${dns_validation_file_tmp}" 1>>"${dev_null}" || true
  info "Waiting for the test pod to come up"
  kubectl wait --for=condition=Ready --timeout=300s pod dns-validation-istio-test
  info "Test pod is ready"

  while IFS=$'\n' read -r subdomain_resolve_map_iter
  do
    domain=$(echo "${subdomain_resolve_map_iter}" | cut -d'=' -f1)
    domain_ip=$(echo "${subdomain_resolve_map_iter}" | cut -d'=' -f2)
    pod_domain_ip=""
    for ((i=0;i<"${resolution_try}";i++))
    do
      pod_domain_ip=$(kubectl exec dns-validation-istio-test -n default -- getent ahosts  "${domain}" 2>>"${dev_null}" | awk '{print $1}' | sort | uniq | tr '\n' ',' | sed 's/,$//g')
      [[ -n "${pod_domain_ip}" ]] && break
    done
    if [[ -z "${pod_domain_ip}" ]]
    then
      error_messages+=("No IP Address found for ${domain} from container. Resolved to ${domain_ip} from host")
    elif [[ "$pod_domain_ip" != "$domain_ip" ]]
    then
      warn_messages+=("Resolved ip(s) for ${domain} not same when tried from host(${domain_ip}) and container(${pod_domain_ip})")
    fi
  done < /tmp/subdomain_resolve_map.txt
  if [[ ${#warn_messages[@]} != 0 ]] ; then
    warn_output "Discrepancy found in DNS config from container" "${warn_messages[@]}"
  fi
  if [[ ${#error_messages[@]} != 0 ]] ; then
    failure_output "DNS Validation Failed from container" "${error_messages[@]}"
  fi

  info "Deleting the test pod after validation"
  kubectl delete pod dns-validation-istio-test -n default 1>>"${dev_null}" || true
}


function download_istio_helm_chart() {
    local docker_registry_password_name="docker-registry-credentials"
    local OFFLINE_TMP_FOLDER="/tmp"
    chartVersion="$(get_chart_version_from_manifest "rancher-istio" "vendor")"
    tmp_helm_charts_dir=$(mktemp -d -p "${OFFLINE_TMP_FOLDER}" -t sf-fabric-helm-charts.XXXXXXXXXX)

    declare -a charts=(
      "vendor/rancher-istio ${chartVersion}"
    )
    if [[ "${INSTALL_TYPE}" == "offline"  ]]; then
      helm_registry_url="${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}"
      helm_registry_password=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.password\}\} | base64 -d)
      helm_registry_username=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.username\}\} | base64 -d)
    else
      helm_registry_url="$(get_mandatory_json_value ".registries.helm.url")"
      helm_registry_username="$(get_json_value_or_return_default_value ".registries.helm.username" "-")"
      helm_registry_password="$(get_json_value_or_return_default_value ".registries.helm.password" "-")"
    fi

    download_fabric_helm_charts "${helm_registry_url}" "${helm_registry_username}" "${helm_registry_password}" "${tmp_helm_charts_dir}" "${charts[@]}"
    export tmp_helm_charts_dir

}
function download_istio_configure_chart() {
    local docker_registry_password_name="docker-registry-credentials"
    local OFFLINE_TMP_FOLDER="/tmp"
    chartVersion="$(get_chart_version_from_manifest "istio-configure" "uipath")"
    tmp_helm_charts_dir=$(mktemp -d -p "${OFFLINE_TMP_FOLDER}" -t sf-fabric-helm-charts.XXXXXXXXXX)

    declare -a charts=(
      "helm/istio-configure ${chartVersion}"
    )
    if [[ "${INSTALL_TYPE}" == "offline" ]]; then
      helm_registry_url="${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}"
      helm_registry_password=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.password\}\} | base64 -d)
      helm_registry_username=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.username\}\} | base64 -d)
    else
      helm_registry_url="$(get_mandatory_json_value ".registries.helm.url")"
      helm_registry_username="$(get_json_value_or_return_default_value ".registries.helm.username" "-")"
      helm_registry_password="$(get_json_value_or_return_default_value ".registries.helm.password" "-")"
    fi

    download_fabric_helm_charts "${helm_registry_url}" "${helm_registry_username}" "${helm_registry_password}" "${tmp_helm_charts_dir}" "${charts[@]}"
    export tmp_helm_charts_dir

}

function test_istio_installation() {
  local fqdn
  [[ ${ALTERNATIVE_FQDN} == https://* ]] && fqdn="${ALTERNATIVE_FQDN}" || fqdn="https://${ALTERNATIVE_FQDN}"
  info "Going to test istio on endpoint ${fqdn}"

  retry=0
  until [ "$retry" -ge 30 ]
  do
    response=$(curl -i -k -sS -m 10 "${fqdn}") || warn "Istio test failed, Retry no ${retry}...."
    grep -q "istio" <<< "${response}" && info "Istio test passed" && return
    retry=$((retry+1))
    sleep 10
  done
  error_with_code "11023" "${retry}"
}

