#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

LANG_FABRIC_FAILED_PREFIX="[ManagementTools] [Failed]"
LANG_STORAGE_SETUP_MESSAGE="Installing and configuring storage"

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/longhornutils.sh"


function load_lh_variables() {
  PROFILE="$(get_mandatory_json_value ".profile")"
  FAULT_TOLERANCE="$(get_json_value_or_return_default_value ".fault_tolerance" "1")"

  [[ "${PROFILE}" == "ha" && "$FAULT_TOLERANCE" -lt 1 ]] && error "Unsupported Fault tolerance value for HA profile. Expected: '>=1' Current: '${FAULT_TOLERANCE}'"
  [[ "${PROFILE}" != "ha" && "$FAULT_TOLERANCE" -ne 0 ]] && info "Setting Fault tolerance to '0' for non-HA profile" && FAULT_TOLERANCE=0

  ZONE_RESILIENCE="$(get_json_value_or_return_default_value ".zone_resilience" "false")"
  MOUNT_POINT="$(get_mandatory_json_value ".infra.longhorn.mount")"

  STALE_REPLICA_TIMEOUT="$(get_mandatory_json_value ".infra.longhorn.staleReplicaTimeout")"
  REPLICAS="$(get_mandatory_json_value ".infra.longhorn.replicas")"
  LH_RECURRING_SNAPSHOT_SCHEDULE="$(get_mandatory_json_value ".infra.longhorn.recurring_snapshot_schedule")"
  DOCKER_REGISTRY_LOCAL_USERNAME="$(get_mandatory_json_value ".infra.docker_registry.username")"
  DOCKER_REGISTRY_LOCAL_PASSWORD="$(get_mandatory_json_value ".infra.docker_registry.password")"
}

function push_longhorn_images_to_docker_registry() {
  local containerdSocket="/run/k3s/containerd/containerd.sock"
  local agentImageDirectory="/var/lib/rancher/rke2/agent/images"
  local ctr_ns="k8s.io"
  local docker_ns="docker-registry"
  local image_file="../Infra_Installer/longhorn/longhorn-image-list"
  local max_attempt=10
  local retry
  local DOCKER_REGISTRY_LOCAL_USERNAME="$(get_mandatory_json_value ".infra.docker_registry.username")"
  local DOCKER_REGISTRY_LOCAL_PASSWORD="$(get_mandatory_json_value ".infra.docker_registry.password")"

  completed=1
  retry=0
  # Wait for docker-registry pod to become ready
  # Because of rke restart, docker-registry pod may get into multi volume-attach error
  while [ "${retry}" -lt "${max_attempt}" ]; do
    kubectl -n "${docker_ns}" wait --for=condition=Ready -l "app=docker-registry" pods --timeout=30s || { sleep 30 && retry=$((retry+1)) && info "Waiting for docker-registry pods to become ready.. attempt ${retry}/$max_attempt}" && continue; }
    completed=0 && break
  done
  [[ "${completed}" -eq 1 ]] && error "Timed out waiting for docker-registry pods to become ready"

  completed=0
  while read -r img; do
    image_name=$(echo "${img}" |  cut -d':' -f1 | tr '/' '_')
    image_tag=$(echo "${img}" | awk -F'/' '{print $NF}' | cut -d':' -f2 | tr '.' '_')
    file_name="${image_name}_${image_tag}.tar"

    retry=0
    completed=0
    while [ "$retry" -lt "$max_attempt" ]; do
      retry=$((retry+1))

      #Remove image from containered before update it to avoid missing layers issue
      remove_image_from_containerd_cache "docker.io/${img}"
      remove_image_from_containerd_cache "${DOCKER_REGISTRY_URL}/${img}"
      ctr -a "${containerdSocket}" -n "${ctr_ns}" image import "${agentImageDirectory}"/"${file_name}" 2>>"${dev_null}" || true
      ctr -a "${containerdSocket}" -n "${ctr_ns}" images tag "docker.io/${img}" "${DOCKER_REGISTRY_URL}/${img}" 2>>"${dev_null}" || true
      ctr -a "${containerdSocket}" -n "${ctr_ns}" images push "${DOCKER_REGISTRY_URL}/${img}" -k -u "${DOCKER_REGISTRY_LOCAL_USERNAME}:${DOCKER_REGISTRY_LOCAL_PASSWORD}" 2>>"${dev_null}" || { info "Image seeding to registry for image:${img} failed.. retry ${retry}/${max_attempt}" && sleep 30 && continue; }
      completed=1 && break
    done
    if [[ "$completed" -eq 0 ]]; then
      error "Image seeding to registry for image:${img} failed"
    fi
  done <"${image_file}"
}

function run_longhorn_image_copier() {
  local ns="kube-system"
  local cm_script_name="longhorn-image-copy-cm"
  local cm_image_data_name="longhorn-image-copy-data"
  local image_pull_secret="longhorn-image-cred"
  local ds_name="longhorn-image-copier"
  local registry_username
  local registry_password
  local DOCKER_REGISTRY_LOCAL_USERNAME="$(get_mandatory_json_value ".infra.docker_registry.username")"
  local DOCKER_REGISTRY_LOCAL_PASSWORD="$(get_mandatory_json_value ".infra.docker_registry.password")"

  registry_username=$(echo "${DOCKER_REGISTRY_LOCAL_USERNAME}" | base64 -w 0 |  sed -e 's/[[:space:]]//g')
  registry_password=$(echo "${DOCKER_REGISTRY_LOCAL_PASSWORD}" | base64 -w 0 |  sed -e 's/[[:space:]]//g')

  #Cleanup old resources
  kubectl -n "${ns}" delete secret "${image_pull_secret}" || true
  kubectl -n "${ns}" delete cm "${cm_script_name}" "${cm_image_data_name}" || true
  kubectl -n "${ns}" delete ds "${ds_name}" || true
  kubectl -n "${ns}" rollout status daemonset "${ds_name}" || true

  #Deploy longhorn image copier
  rm -rf ../Infra_Installer/longhorn/longhorn-image-copy-cred-tmp.yaml
  cp ../Infra_Installer/longhorn/longhorn-image-copy-cred.yaml ../Infra_Installer/longhorn/longhorn-image-copy-cred-tmp.yaml
  sed -i \
    -e "s;##NAME##;${image_pull_secret};g"  \
    -e "s;##USERNAME##;${registry_username};g"  \
    -e "s;##PASSWORD##;${registry_password};g"  \
    ../Infra_Installer/longhorn/longhorn-image-copy-cred-tmp.yaml
  kubectl -n "${ns}" apply -f ../Infra_Installer/longhorn/longhorn-image-copy-cred-tmp.yaml
  kubectl -n "${ns}" create configmap "${cm_image_data_name}" --from-file=../Infra_Installer/longhorn/longhorn-image-list
  kubectl -n "${ns}" create configmap "${cm_script_name}" --from-file=copy-images.sh=../Infra_Installer/longhorn/longhorn-image-copy.sh

  rm -f ../Infra_Installer/longhorn/longhorn-image-copy-ds-tmp.yaml
  cp ../Infra_Installer/longhorn/longhorn-image-copy-ds.yaml ../Infra_Installer/longhorn/longhorn-image-copy-ds-tmp.yaml
  sed -i  \
    -e "s;##REGISTRY_NAME_PLACEHOLDER##;${DOCKER_REGISTRY_URL};g"  \
    -e "s;##IMAGE_COPY_LIST##;${cm_image_data_name};g"  \
    -e "s;##COPIER_SCRIPT_CM_NAME##;${cm_script_name};g"  \
    -e "s;##IMAGE_CRED_NAME##;${image_pull_secret};g"  \
         ../Infra_Installer/longhorn/longhorn-image-copy-ds-tmp.yaml

  kubectl -n "${ns}" apply -f ../Infra_Installer/longhorn/longhorn-image-copy-ds-tmp.yaml
  kubectl -n "${ns}" rollout status daemonset "${ds_name}"

  POD_COUNT=$(kubectl get nodes -o json | jq -r '.items | length')
  local COUNT=0
  local max_attempt=10
  local attempt=0
  local ret_code

  ret_code=1
  while [ "${attempt}" -lt "${max_attempt}" ]; do
    kubectl -n "${ns}" wait --for=condition=Ready -l "name=${ds_name}" pods --timeout=30s || { sleep 3 && attempt=$((attempt+1)) && info "Waiting for longhorn image copier pods to become ready.. attempt ${attempt}/$max_attempt}" && continue; }
    ret_code=0 && break
  done
  [[ "${ret_code}" -eq 1 ]] && error "Timed out waiting for $ds_name pods to become ready"

  # Loop through each pod and check if the message is in the logs
  POD_NAMES=$(kubectl -n "${ns}" get pods -l name="${ds_name}" -o json | jq -r '.items[].metadata.name')
  info "${ds_name} daemonset pods are ready"
  for pod in ${POD_NAMES}; do
    attempt=0
    while (( attempt < max_attempt )); do
      attempt=$(( attempt + 1 ))
      if kubectl -n "${ns}" logs "${pod}" |grep "Image copy completed"; then
        COUNT=$((COUNT+1))
        info "pod ${pod} has been completed"
        break
      else
        info "pod ${pod} did not finish its execution in ... ${attempt}/${max_attempt}"
        sleep 60
      fi
    done
  done

  if [ "${COUNT}" == "${POD_COUNT}" ]; then
    kubectl -n "${ns}" delete ds "${ds_name}"
    kubectl -n "${ns}" delete cm "${cm_script_name}" "${cm_image_data_name}"
    kubectl -n "${ns}" delete secret "${image_pull_secret}"

    info "Longhorn images have been copied to all the nodes successfully"
  else
    error "Copying Longhorn images failed on $((POD_COUNT-COUNT)) nodes. Please re-run upgrade"
  fi
}
function longhorn_prereq(){
  info "Validate mount propagation..."
  local try=0
  local maxtry=30
  local longhorn_test_ds_tmp="../Infra_Installer/longhorn/tests/longhorn_test_ds_tmp.yaml"
  local LANG_INFRA_FAILED_PREFIX="[Infrastructure] [Failed]"
  local LANG_STORAGE_SETUP_MESSAGE="Installing and configuring storage"

  \cp "../Infra_Installer/longhorn/tests/longhorn_test_ds.yaml" "${longhorn_test_ds_tmp}"

  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" "${longhorn_test_ds_tmp}"
  else
    sed -i "s;##REGISTRY_PLACEHOLDER##;;" "${longhorn_test_ds_tmp}"
  fi

  kubectl create namespace longhorn-system --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"
  while ! kubectl get namespaces | grep -q "longhorn-system" && (( try != maxtry )) ; do
    try=$((try+1))
    info "Waiting for 'longhorn-system' namespace to be ready...${try}/${maxtry}" &&  sleep 10;
  done

  kubectl get namespaces | grep -q "longhorn-system" || error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}"
  kubectl get namespaces | grep -q "longhorn-system" || error_with_code "11006"

  # Label longhorn-system namespace, this is required in longhorn networkpolicy
  kubectl label namespace longhorn-system name=longhorn-system &>>"${dev_null}" || true

  MOUNTED=$(../Infra_Installer/longhorn/environment_check.sh \
    | grep 'MountPropagation is enabled!')

  [[ -n ${MOUNTED} ]] || {
    error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
    error_with_code "11007";
  }
  info "Validation complete..."
}

function patch_existing_lh_volumes() {
  local spec_name="$1"
  local spec_value="$2"
  # patch existing volume objects with new stale replica timeout
  volume_objects=$(kubectl -n longhorn-system get volumes.longhorn.io --no-headers -o custom-columns=Name:.metadata.name)
  for vol in $volume_objects; do
    kubectl -n longhorn-system patch volumes.longhorn.io "$vol" --type merge -p "{\"spec\" : { \"${spec_name}\": ${spec_value} } }" || true
  done
}

function create_longhorn_stale_folder_cleanup_cronjob() {
   cp ../Infra_Installer/longhorn/longhorn-replica-folder-cleanup.yaml ../Infra_Installer/longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;__REGISTRY_PLACEHOLDER__;${DOCKER_REGISTRY_URL}/;" ../Infra_Installer/longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  else
    sed -i "s;__REGISTRY_PLACEHOLDER__;;" ../Infra_Installer/longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  fi
  sed -i "s;##LONGHORN_DATADISK_PLACEHOLDER##;${MOUNT_POINT}/;" ../Infra_Installer/longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  sed -i "s;__INSTALLER_PATH__;${PATH};g" ../Infra_Installer/longhorn/longhorn-replica-folder-cleanup-tmp.yaml
  kubectl apply -f ../Infra_Installer/longhorn/longhorn-replica-folder-cleanup-tmp.yaml 1>>"${dev_null}" || true
}

function create_longhorn_loop_device_cleaner_daemonset() {
  cp ../Infra_Installer/longhorn/longhorn-loop-device-cleaner.yaml ../Infra_Installer/longhorn/longhorn-loop-device-cleaner-tmp.yaml
  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;__REGISTRY_PLACEHOLDER__;${DOCKER_REGISTRY_URL}/;" ../Infra_Installer/longhorn/longhorn-loop-device-cleaner-tmp.yaml
  else
    sed -i "s;__REGISTRY_PLACEHOLDER__;;" ../Infra_Installer/longhorn/longhorn-loop-device-cleaner-tmp.yaml
  fi
  kubectl apply -f ../Infra_Installer/longhorn/longhorn-loop-device-cleaner-tmp.yaml 1>>"${dev_null}" || true
}

function patch_existing_volumes_stale_timeout() {
  # patch existing volume objects with new stale replica timeout
  volume_objects=$(kubectl -n longhorn-system get volumes.longhorn.io --no-headers -o custom-columns=Name:.metadata.name)
  for vol in $volume_objects;
  do
    kubectl -n longhorn-system patch volumes.longhorn.io "$vol" --type merge -p "{\"spec\" : { \"staleReplicaTimeout\": ${STALE_REPLICA_TIMEOUT} } }" || true
  done
}

function install_longhorn(){

  LONGHORN_VERSION="$(get_chart_version_from_manifest "longhorn" "vendor")"
  HA_REPLICAS="$(get_mandatory_json_value ".infra.longhorn.ha_replicas")"
  ZONE_RESILIENCE="$(get_json_value_or_return_default_value ".zone_resilience" "false")"
  MOUNT_POINT="$(get_mandatory_json_value ".infra.longhorn.mount")"
  local LANG_INFRA_FAILED_PREFIX="[Infrastructure] [Failed]"
  local LANG_STORAGE_SETUP_MESSAGE="Installing and configuring storage"
  local docker_registry_password_name="docker-registry-credentials"

  declare -a charts=(
    "vendor/longhorn ${LONGHORN_VERSION}"
  )

if [[ ${DO_UPGRADE} == "1" ]]; then
  #shellcheck disable=SC2154
  if [[ "${INSTALL_TYPE}" == "offline"  ]]; then
      helm_registry_url="${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}"
      helm_registry_password=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.password\}\} | base64 -d)
      helm_registry_username=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.username\}\} | base64 -d)
    else
      helm_registry_url="$(get_mandatory_json_value ".registries.helm.url")"
      helm_registry_username="$(get_json_value_or_return_default_value ".registries.helm.username" "-")"
      helm_registry_password="$(get_json_value_or_return_default_value ".registries.helm.password" "-")"
      download_fabric_helm_charts "${helm_registry_url}" "${helm_registry_username}" "${helm_registry_password}" "${TMP_HELM_CHARTS_DIR}" "${charts[@]}"
    fi
  
fi
  

  if [[ "${DO_UPGRADE}" == "1" ]]; then
    # Set concurrent-automatic-engine-upgrade-per-node-limit to 1 for upgrading engine image automatically
    kubectl patch -n longhorn-system setting concurrent-automatic-engine-upgrade-per-node-limit --type=json -p='[{"op":"replace", "path":"/value", "value":"5"}]'
  fi

  [[ ${INSTALL_TYPE} == "online" ]] && LONGHORN_ARGS+=(--set global.cattle.systemDefaultRegistry="${DOCKER_REGISTRY_URL}")
  [[ -n ${LONGHORN_VERSION} ]] && LONGHORN_ARGS+=(--version="${LONGHORN_VERSION}")

  LONGHORN_ARGS+=(--set defaultSettings.autoDeletePodWhenVolumeDetachedUnexpectedly=true)
  if [[ $PROFILE != "ha" ]]; then
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedReplicaManagerCPU=1)
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedEngineManagerCPU=1)
  else
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedReplicaManagerCPU=10)
    LONGHORN_ARGS+=(--set defaultSettings.guaranteedEngineManagerCPU=10)
  fi;
  LONGHORN_ARGS+=(--set defaultSettings.storageMinimalAvailablePercentage=5)
  LONGHORN_ARGS+=(--set defaultSettings.allowNodeDrainWithLastHealthyReplica=false)
  LONGHORN_ARGS+=(--set defaultSettings.nodeDrainPolicy=block-if-contains-last-replica)
  LONGHORN_ARGS+=(--set defaultSettings.defaultDataLocality=best-effort)
  LONGHORN_ARGS+=(--set defaultSettings.nodeDownPodDeletionPolicy=delete-both-statefulset-and-deployment-pod)

  LONGHORN_ARGS+=(--set defaultSettings.priorityClass=system-cluster-critical)
  LONGHORN_ARGS+=(--set longhornManager.priorityClass=system-cluster-critical)
  LONGHORN_ARGS+=(--set longhornDriver.priorityClass=system-cluster-critical)
  LONGHORN_ARGS+=(--set longhornUI.priorityClass=system-cluster-critical)
  LONGHORN_ARGS+=(--set enablePSP=false)

  if kubectl -n longhorn-system get cm longhorn-default-setting &>>"${dev_null}";
  then
    OVERPROVISIONING_PERCENTAGE=$(kubectl -n longhorn-system get cm longhorn-default-setting -o jsonpath='{.data.default-setting\.yaml}' | grep storage-over-provisioning-percentage: | cut -d ':' -f2 | xargs)
    if [[ -n "${OVERPROVISIONING_PERCENTAGE}" ]];
    then
      info "Longhorn default settings found with storageOverProvisioningPercentage set to ${OVERPROVISIONING_PERCENTAGE}"
      LONGHORN_ARGS+=(--set defaultSettings.storageOverProvisioningPercentage="${OVERPROVISIONING_PERCENTAGE}")
    fi
  else
    info "Longhorn default settings not found, setting storageOverProvisioningPercentage to 100%"
    LONGHORN_ARGS+=(--set defaultSettings.storageOverProvisioningPercentage=100)
  fi

  REPLICAS_BACKUP_ONE=1
  if [[ $INSTALL_MULTINODE == "true" ]]; then
    REPLICAS=${HA_REPLICAS}
    REPLICAS_BACKUP_ONE=2
    LONGHORN_ARGS+=(--set defaultSettings.replicaAutoBalance=best-effort)
    if [[ $ZONE_RESILIENCE == "true" ]]; then
      LONGHORN_ARGS+=(--set defaultSettings.replicaZoneSoftAntiAffinity=false)
    fi
  else
    # reset csi pods to 1 replica for single node
    LONGHORN_ARGS+=(--set csi.attacherReplicaCount=1)
    LONGHORN_ARGS+=(--set csi.provisionerReplicaCount=1)
    LONGHORN_ARGS+=(--set csi.resizerReplicaCount=1)
    LONGHORN_ARGS+=(--set csi.snapshotterReplicaCount=1)
  fi

  LONGHORN_ARGS+=(--set persistence.defaultClassReplicaCount="1")
  LONGHORN_ARGS+=(--set defaultSettings.defaultReplicaCount="1")
  #defaultDataPath can only be set at installation time, not upgrade
  [[ -n ${MOUNT_POINT} ]] && LONGHORN_ARGS+=(--set defaultSettings.defaultDataPath="${MOUNT_POINT}")

  info "Installing longhorn with arguments ${LONGHORN_ARGS[*]}"

  helm upgrade --install --atomic -f ../Infra_Installer/longhorn/longhorn-tolerations-values.yaml longhorn "${TMP_HELM_CHARTS_DIR}"/longhorn/ \
    --wait --timeout 10m \
    --namespace longhorn-system \
    --create-namespace \
    "${LONGHORN_ARGS[@]}" &>>"${dev_null}" || {
      error_without_exit "${LANG_INFRA_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}";
      error_with_code "11008";
    }


  # create default backup storage class
  rm -f ../Infra_Installer/longhorn/longhorn-storage-backup-tmp.yaml &>>"${dev_null}"
  cp ../Infra_Installer/longhorn/longhorn-storage-backup.yaml ../Infra_Installer/longhorn/longhorn-storage-backup-tmp.yaml
  sed -i "s/REPLICA_PLACEHOLDER/${REPLICAS}/" ../Infra_Installer/longhorn/longhorn-storage-backup-tmp.yaml
  sed -i "s/STALE_REPLICA_TIMEOUT_PLACEHOLDER/${STALE_REPLICA_TIMEOUT}/" ../Infra_Installer/longhorn/longhorn-storage-backup-tmp.yaml
  # Delete older storageclass definitions so that new parameters are applied
  kubectl delete -f ../Infra_Installer/longhorn/longhorn-storage-backup-tmp.yaml 1>>"${dev_null}" || true
  kubectl apply -f ../Infra_Installer/longhorn/longhorn-storage-backup-tmp.yaml 1>>"${dev_null}"
  # create backup storage class with single replica

  rm -f ../Infra_Installer/longhorn/longhorn-storage-backup-one-replica-tmp.yaml &>>"${dev_null}"
  cp ../Infra_Installer/longhorn/longhorn-storage-backup-one-replica.yaml ../Infra_Installer/longhorn/longhorn-storage-backup-one-replica-tmp.yaml
  sed -i "s/REPLICA_PLACEHOLDER/${REPLICAS_BACKUP_ONE}/" ../Infra_Installer/longhorn/longhorn-storage-backup-one-replica-tmp.yaml
  sed -i "s/STALE_REPLICA_TIMEOUT_PLACEHOLDER/${STALE_REPLICA_TIMEOUT}/" ../Infra_Installer/longhorn/longhorn-storage-backup-one-replica-tmp.yaml
  # Delete older storageclass definitions so that new parameters are applied
  kubectl delete -f ../Infra_Installer/longhorn/longhorn-storage-backup-one-replica-tmp.yaml 1>>"${dev_null}" || true
  kubectl apply -f ../Infra_Installer/longhorn/longhorn-storage-backup-one-replica-tmp.yaml 1>>"${dev_null}"

  rm -f ../Infra_Installer/longhorn/longhorn-storage-tmp.yaml &>>"${dev_null}"
  cp ../Infra_Installer/longhorn/longhorn-storage.yaml ../Infra_Installer/longhorn/longhorn-storage-tmp.yaml
  sed -i "s/STALE_REPLICA_TIMEOUT_PLACEHOLDER/${STALE_REPLICA_TIMEOUT}/" ../Infra_Installer/longhorn/longhorn-storage-tmp.yaml
  # Delete older storageclass definitions so that new parameters are applied
  kubectl delete -f ../Infra_Installer/longhorn/longhorn-storage-tmp.yaml 1>>"${dev_null}" || true
  kubectl apply -f ../Infra_Installer/longhorn/longhorn-storage-tmp.yaml 1>>"${dev_null}"

  patch_existing_volumes_stale_timeout

  # create longhorn networkpolicy to block unauthenticated access
  USE_CILIUM=1
  if [[ "${USE_CILIUM}" -gt 0 ]]; then
    if kubectl -n longhorn-system get networkpolicy longhorn-networkpolicy &>>"${dev_null}";
    then
      # Delete the older networkpolicy
      kubectl -n longhorn-system delete networkpolicy longhorn-networkpolicy
    fi
  else
    kubectl apply -f ../Infra_Installer/longhorn/longhorn-networkpolicy.yaml &>>"${dev_null}"
  fi

  # create longhorn pdb for csi-attacher
  kubectl apply -f ../Infra_Installer/longhorn/longhorn-attacher-pdb.yaml &>>"${dev_null}"

  # longhorn conversion webhook may delay updating TLS certificate, longhorn-webhook-tls
  sleep 120
  # create longhorn recurring snapshot if backup not enabled
  if ! kubectl -n longhorn-system get cronjob uipath-backup &>>"${dev_null}"; then
    rm -f ../Infra_Installer/longhorn/longhorn-recurring-snapshot-tmp.yaml &>>"${dev_null}" || true
    cp ../Infra_Installer/longhorn/longhorn-recurring-snapshot.yaml ../Infra_Installer/longhorn/longhorn-recurring-snapshot-tmp.yaml
    sed -i "s/SCHEDULE_PLACEHOLDER/${LH_RECURRING_SNAPSHOT_SCHEDULE}/g" ../Infra_Installer/longhorn/longhorn-recurring-snapshot-tmp.yaml
    kubectl apply -f ../Infra_Installer/longhorn/longhorn-recurring-snapshot-tmp.yaml &>>"${dev_null}"
  fi

  if ! kubectl -n longhorn-system get cronjob uipath-snapshot-purge &>>"${dev_null}"; then
    rm -f ../Infra_Installer/longhorn/longhorn-purge-system-snapshot-tmp.yaml &>>"${dev_null}" || true
    cp ../Infra_Installer/longhorn/longhorn-purge-system-snapshot.yaml ../Infra_Installer/longhorn/longhorn-purge-system-snapshot-tmp.yaml
    sed -i "s/SCHEDULE_PLACEHOLDER/${LH_RECURRING_SNAPSHOT_SCHEDULE}/g" ../Infra_Installer/longhorn/longhorn-purge-system-snapshot-tmp.yaml
    kubectl apply -f ../Infra_Installer/longhorn/longhorn-purge-system-snapshot-tmp.yaml &>>"${dev_null}"

  fi

  create_longhorn_stale_folder_cleanup_cronjob
  create_longhorn_loop_device_cleaner_daemonset


  # update longhorn csi-provisioner for 1.2.0 release https://github.com/longhorn/longhorn/issues/2964
  if [[ "${LONGHORN_VERSION}" == "1.2.0" ]]; then
    update_csi_provisioner
  fi
  info "Longhorn Installation Success"
}



function update_csi_provisioner() {
  local try=0
  local maxtry=10
  local csi_provisioner_exist=0

  while ((try < maxtry)); do
    result=$(kubectl -n longhorn-system get deploy csi-provisioner) || true
    if [[ -n "${result}" ]]; then
      csi_provisioner_exist=1
      break
    fi

    try=$((try + 1))
    info "Waiting for csi-provisioner deployment to come up. ${try}/${maxtry}..."
    sleep 5
  done

  if ((csi_provisioner_exist == 0)); then
    error_with_code "11008"
  fi

  local container_data
  container_data=$(kubectl -n longhorn-system get deploy csi-provisioner -o json | jq -c ".spec.template.spec.containers[]| select( .name == \"csi-provisioner\").args[.args| length] |= . + \"--default-fstype=ext4\"")
  kubectl -n longhorn-system patch deployment csi-provisioner --patch "{\"spec\": {\"template\": {\"spec\": {\"containers\": [${container_data}]}}}}"
}

function test_longhorn() {
  #shellcheck disable=SC2086
  if ! ../Infra_Installer/longhorn/longhorn-readiness-check.sh; then
    error_without_exit "${LANG_FABRIC_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}"
    error_with_code "11042"
  fi

  local pod_with_pvc_tmp="../Infra_Installer/longhorn/tests/pod_with_pvc-tmp.yaml"

  # create a copy of original file to avoid modifying the original file
  \cp "../Infra_Installer/longhorn/tests/pod_with_pvc.yaml" "${pod_with_pvc_tmp}"

  # if INSTALL_TYPE is offline, then the registry image is already present on the machine
  # if an external registry is provided use it instead of the local image
  # if INSTALL_TYPE is online, then images need to be pulled from DOCKER_REGISTRY_URL
  if [[ ${INSTALL_TYPE} == "online" ]]; then
    sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" "${pod_with_pvc_tmp}"
  else
    sed -i "s;##REGISTRY_PLACEHOLDER##;;" "${pod_with_pvc_tmp}"
  fi


  kubectl delete -n longhorn-system -f "${pod_with_pvc_tmp}" --ignore-not-found=true &>>"${dev_null}"
  kubectl create -n longhorn-system -f "${pod_with_pvc_tmp}" &>>"${dev_null}"

  retry=0
  until [ "$retry" -ge 5 ]; do
    kubectl wait -n longhorn-system --for=condition=Ready --timeout=2m pods volume-test &>>"${dev_null}" && break
    kubectl describe pods -n longhorn-system volume-test

    info "Longhorn test failed, Retry no ${retry}...."
    kubectl delete -n longhorn-system -f "${pod_with_pvc_tmp}" --ignore-not-found=true &>>"${dev_null}"
    kubectl create -n longhorn-system -f "${pod_with_pvc_tmp}" &>>"${dev_null}"
    retry=$((retry + 1))
  done

  POD=$(kubectl get pods -n longhorn-system volume-test |
    grep volume-test |
    grep Running |
    grep "1/1")

  BOUND=$(kubectl get pvc -n longhorn-system -A |
    grep longhorn-volv-pvc |
    grep Bound)

  kubectl delete -n longhorn-system -f "${pod_with_pvc_tmp}" &>>"${dev_null}"

  [[ -n ${POD} ]] || {
    error_without_exit "${LANG_FABRIC_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}"
    error_with_code "11009"
  }

  [[ -n ${BOUND} ]] || {
    error_without_exit "${LANG_FABRIC_FAILED_PREFIX} ${LANG_STORAGE_SETUP_MESSAGE}"
    error_with_code "11010"
  }
}