#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/k8sutils.sh"

SERVICE_NAME=rke2-server.service
[[ "${NODE_ROLE}" == "agent" ]] && SERVICE_NAME=rke2-agent.service
POST_INFRA_UPGRADE_LOCK_FILE_PATH="${SUC_RKE_RESTART_STATE_FILENAME}.postInfra"

function is_kube_proxy_healthy() {
  local kube_dns_server
  local istio_ingress_gw_svc_cluster_ip

  crictl -r unix:///run/k3s/containerd/containerd.sock ps | grep ' kube-proxy ' || return 1
  kube_dns_server=$(kubectl -n kube-system get service rke2-coredns-rke2-coredns -o json | jq -r '.spec.clusterIP') || return 1
  [[ "$kube_dns_server" == "null" || -z "$kube_dns_server" ]] && return 1
  istio_ingress_gw_svc_cluster_ip=$(dig @"${kube_dns_server}" +short istio-ingressgateway.istio-system.svc.cluster.local) || return 1
  [[ -z "$istio_ingress_gw_svc_cluster_ip" ]] && return 1
  for ((i = 0; i < 5; i++)); do
    curl -s -f -m 5 "http://${istio_ingress_gw_svc_cluster_ip}:15021/healthz/ready" || return 1
  done
}

function restart_rke_background() {
  setsid bash -xc '
    echo "Stopping service: '"${SERVICE_NAME}"'"
    systemctl stop '"${SERVICE_NAME}"'
    echo "Creaing lock file: '"${POST_INFRA_UPGRADE_LOCK_FILE_PATH}"'"
    touch '"${POST_INFRA_UPGRADE_LOCK_FILE_PATH}"'
    echo "Starting service: '"${SERVICE_NAME}"'"
    systemctl start '"${SERVICE_NAME}"'
  ' >/tmp/post_infra_rke_restart.log 2>&1
}

function wait_for_kube_proxy_healthy() {
  local try="${1:-10}"
  local is_kubeproxy_healthy=false

  for ((i = 1; i <= "${try}"; i++)); do
    is_kube_proxy_healthy && {
      is_kubeproxy_healthy=true
      break
    }
    [[ "$i" -ne "$try" ]] && {
      info "Waiting for kube proxy to become healthy"
      sleep 10
    }
  done

  [[ "$is_kubeproxy_healthy" == "true" ]] && return 0
  return 1
}

function remove_kube_proxy() {
  local kube_proxy_json
  local kube_proxy_container_id
  local kube_proxy_pod_id

  kube_proxy_json=$(crictl -r unix:///run/k3s/containerd/containerd.sock ps --name kube-proxy -o json)
  for kube_proxy_container_id in $(echo "${kube_proxy_json}" | jq -r '.containers[].id'); do
    crictl -r unix:///run/k3s/containerd/containerd.sock rm --force "${kube_proxy_container_id}" || true
  done

  for kube_proxy_pod_id in $(echo "${kube_proxy_json}" | jq -r '.containers[].podSandboxId'); do
    crictl -r unix:///run/k3s/containerd/containerd.sock rmp --force "${kube_proxy_pod_id}" || true
  done
}

function main() {
  # Remove kube proxy container as updated / healthy rke service should create it back after deletion
  # This is required as original restart effect may take some time for kube-proxy to be come unhealthy
  # causing the below kube proxy health check pass and but kube proxy fails later on
  remove_kube_proxy
  wait_for_kube_proxy_healthy || {
    [[ -f "${POST_INFRA_UPGRADE_LOCK_FILE_PATH}" && "${ENABLE_SUC_UPGRADE_FF}" == "1" ]] && {
      error_with_code "14034"
    }
    # Run restart in background
    restart_rke_background
    # Check healthy status of kube-proxy
    wait_for_kube_proxy_healthy "30" || error "Kube Proxy is not ready even after restart"
  }
  info "Kube Proxy is ready"
}

main