#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/upgrade_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/argocd_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/cilium_utils.sh"

function main() {
  if [[ "${NODE_ROLE}" != "none" ]]; then
    info "Skipping rancher server upgrade run from secondary nodes"
    return
   fi
  argo_app_name="fabric-installer"
  argo_namespace="argocd"
  argo_app=$(kubectl get application "${argo_app_name}" -n "${argo_namespace}" --ignore-not-found)
  info "Disabling auto-sync on $argo_app_name"
  if [[ -n "${argo_app}" ]]; then
    kubectl -n argocd patch application "${argo_app_name}" --type=json -p '[{"op":"replace","path":"/spec/syncPolicy/automated/selfHeal","value":false}]' || true
  else
    warn "Failed to find ArgoCD app ${argo_app_name}"
  fi

  if [[ "$(is_cilium_installed)" == "true" ]]; then
    apply_rancher_network_policy
  fi

  # https://github.com/rancher/rke2/issues/2085#issuecomment-963541733
  kubectl delete helmchart -n kube-system rke2-metrics-server

  # startupapicheck is a new component for cert-manager 1.9.1 which needs some parameter override
  image_host=$(kubectl get deployment -n cert-manager cert-manager -o jsonpath='{.spec.template.spec.containers[0].image}' | cut -d/ -f1)
  argocd_fqdn=$(get_cluster_config "ARGOCD_FQDN")
  argocd_nodeport_http=$(get_cluster_config_default "CLUSTER_PORT" 443)
  argocd_bin="${MODULES_DIR}/../bin/argocd"
  login_argocd "${argocd_nodeport_http}" "${argocd_fqdn}" "argocd-admin-password" "argocd" "${argocd_bin}"
  "${argocd_bin}" app set cert-manager -p startupapicheck.image.repository="${image_host}/jetstack/cert-manager-ctl"

  # patch gatekeeper to ensure it does not add 'seccompProfile' which causes below error
  # Forbidden: seccomp may not be set
  patch_argocd_gatekeeper_disable_seccomp

  update_app_version "cert-manager" "1.9.1"
  update_app_version "gatekeeper-crd" "102.1.0-up3.12.0"
  update_app_version "gatekeeper" "102.1.0-up3.12.0"
  update_app_version "rancher-logging-crd" "102.0.1-up3.17.10"
  update_app_version "rancher-logging" "102.0.1-up3.17.10"
  update_app_version "rancher-server" "2.7.6"

  sleep 5

  wait_app_ready "cert-manager" "1.9.1"
  wait_app_ready "gatekeeper-crd" "102.1.0-up3.12.0"
  wait_app_ready "gatekeeper" "102.1.0-up3.12.0"
  wait_app_ready "rancher-logging-crd" "102.0.1-up3.17.10"
  wait_app_ready "rancher-logging" "102.0.1-up3.17.10"
  wait_app_ready "rancher-server" "2.7.6"

  # rancher-webhook sometimes doesn't seem to be updated from 0.2.0 to 0.2.6, manually doing it as per rancher's suggestion
  webhook_image=$(kubectl get deployment -n cattle-system rancher-webhook -o jsonpath='{.spec.template.spec.containers[0].image}' | awk -F ':v' '{print $1":v0.3.5"}')
  kubectl patch deployment  -n cattle-system rancher-webhook   -p "{\"spec\" : {\"template\" : {\"spec\" : {\"containers\":[{\"name\":\"rancher-webhook\",\"image\":\"${webhook_image}\"}]}}}}"
  info "Pods in cattle-system namespace after the patch"
  kubectl get pods -n cattle-system -o wide
}

function wait_app_ready() {
  local name=$1
  local version=$2
  namespace="argocd"

  info "Waiting for ArgoCD app ${name} to be synced with version ${version}"
  local try=0
  local max=10
  while (( try < max ));
  do
    revision=$(kubectl get application "${name}" -n "${namespace}" -o jsonpath='{.status.sync.revision}')
    status=$(kubectl get application "${name}" -n "${namespace}" -o jsonpath='{.status.sync.status}')
    if [[ -n "${revision}" || -n "${status}" ]]; then
      info "[${try}/${max}] ArgoCD app ${name} - Revision: ${revision}, Status: ${status}"
      if [[ "${status}" == "Synced" && "${revision}" == "${version}" ]]; then
        info "ArgoCD App ${name} is ready"
        break
      fi
    else
      warn "Failed to find ArgoCD app ${name}"
    fi
    sleep 30
    try=$((try+1))
  done
}

function update_app_version() {
  local name=$1
  local version=$2
  namespace="argocd"

  info "Updating ArgoCD app ${name} to version ${version}"
  argo_app=$(kubectl get application "${name}" -n "${namespace}" --ignore-not-found)
  if [[ -n "${argo_app}" ]]; then
    kubectl patch application "${name}"\
                              -n "${namespace}" \
                              --type=merge \
                              -p "{\"spec\" : {\"source\" : {\"targetRevision\" : \"${version}\"}}}"
  else
    warn "Failed to find ArgoCD app ${name}"
  fi
}

info "Pre-infra upgrade of Rancher components"
if [[ "${NODE_TYPE}" == "server" ]]; then
  main
else
  info "Skipping pre-upgrade-rancher for non master node."
fi